@@ -3,8 +3,8 @@ mod tag;
 use self::tag::Tag;
 use crate::{SyntaxKind, SyntaxNode, SyntaxToken};
 use std::marker::PhantomData;
-use std::slice;
 use std::sync::atomic::{AtomicU32, Ordering};
+use std::{fmt, slice};
 use text_size::TextRange;
 
 /// `SyntaxTree` owns the syntax tree allocation.
@@ -28,7 +28,7 @@ use text_size::TextRange;
 /// - `[u8]` UTF-8 encoded text
 /// - `[u8]` events
 ///
-/// These are stored contiguously in one memory allocation (`Vec<u8>`).
+/// These are stored contiguously in one memory allocation.
 /// Nodes and tokens are a `u32` index into this allocation.
 /// All numerical types are stored in the target platform’s native endianness.
 ///
@@ -86,7 +86,7 @@ use text_size::TextRange;
 /// we must prohibit a kind of fifteen 1s to avoid ambiguity.
 /// Thus, the highest allowed kind is `0b0111_1111_1111_1110`.
 pub struct SyntaxTree<K> {
-    data: Vec<u8>,
+    data: Box<[u8]>,
     phantom: PhantomData<K>,
 }
 
@@ -114,12 +114,33 @@ static CURRENT_TREE_ID: AtomicU32 = AtomicU32::new(0);
 impl<K: SyntaxKind> SyntaxBuilder<K> {
     /// Constructs a new empty `SyntaxBuilder` with the provided source text.
     pub fn new(text: &str) -> Self {
+        Self::with_capacity(text, 0, 0, 0)
+    }
+
+    /// Constructs a new empty `SyntaxBuilder` with the provided source text
+    /// and room for the specified event counts.
+    ///
+    /// Make sure to benchmark before switching to this method
+    /// because precomputing event counts can be slow,
+    /// even slower than just using [`SyntaxBuilder::new`].
+    pub fn with_capacity(
+        text: &str,
+        start_nodes: usize,
+        add_tokens: usize,
+        finish_nodes: usize,
+    ) -> Self {
         debug_assert!(K::LAST <= Tag::MAX_KIND);
         assert!(text.len() < u32::MAX as usize);
 
         let id = CURRENT_TREE_ID.fetch_add(1, Ordering::SeqCst);
 
-        let mut data = id.to_ne_bytes().to_vec();
+        let mut data = Vec::with_capacity(
+            start_nodes * START_NODE_SIZE as usize
+                + add_tokens * ADD_TOKEN_SIZE as usize
+                + finish_nodes * FINISH_NODE_SIZE as usize,
+        );
+
+        data.extend_from_slice(&id.to_ne_bytes());
         data.extend_from_slice(&(text.len() as u32).to_ne_bytes());
         data.extend_from_slice(text.as_bytes());
 
@@ -138,6 +159,7 @@ impl<K: SyntaxKind> SyntaxBuilder<K> {
     /// # Panics
     ///
     /// - if you have finished creating a root node and try to create another
+    #[inline(always)]
     pub fn start_node(&mut self, kind: K) {
         if self.is_root_set {
             assert_ne!(self.nesting, 0, "root node already created");
@@ -166,6 +188,7 @@ impl<K: SyntaxKind> SyntaxBuilder<K> {
     ///
     /// - if you try to add a token before starting a node
     /// - if the provided range is out of bounds of the original input text
+    #[inline(always)]
     pub fn add_token(&mut self, kind: K, range: TextRange) {
         assert!(self.nesting > 0, "cannot add token before starting node");
         assert!(
@@ -193,6 +216,7 @@ impl<K: SyntaxKind> SyntaxBuilder<K> {
     /// # Panics
     ///
     /// - if all outstanding nodes have already been finished
+    #[inline(always)]
     pub fn finish_node(&mut self) {
         assert!(self.nesting > 0, "no nodes are yet to be finished");
         self.nesting -= 1;
@@ -228,16 +252,15 @@ impl<K: SyntaxKind> SyntaxBuilder<K> {
     /// - if no nodes have been created
     /// - if there are nodes which have not been finished
     pub fn finish(self) -> SyntaxTree<K> {
-        let Self { mut data, is_root_set, current_len: _, start_node_idxs: _, nesting, phantom: _ } =
+        let Self { data, is_root_set, current_len: _, start_node_idxs: _, nesting, phantom: _ } =
             self;
 
         assert!(is_root_set, "no nodes created");
 
         assert_eq!(nesting, 0, "did not finish all nodes ({nesting} unfinished nodes)");
 
-        data.shrink_to_fit();
-
-        SyntaxTree { data, phantom: PhantomData }
+        // into_boxed_slice calls shrink_to_fit for us
+        SyntaxTree { data: data.into_boxed_slice(), phantom: PhantomData }
     }
 
     fn text_len(&self) -> u32 {
@@ -255,11 +278,21 @@ impl<K: SyntaxKind> SyntaxTree<K> {
         SyntaxNode::new(self.root_idx(), self.id())
     }
 
+    /// Returns an iterator over the events stored in this tree.
+    ///
+    /// This method does not compute any more information
+    /// than what is stored in the tree.
+    /// The only difference between the [`Event`]s returned by this method
+    /// and what is stored inside the tree
+    /// is that the events returned by this method are fixed-length and typed,
+    /// while the tree’s internal storage is variable-length and untyped.
+    pub fn events(&self) -> impl Iterator<Item = Event<K>> + '_ {
+        Events { idx: self.root_idx(), tree: self }
+    }
+
     pub(crate) fn root_idx(&self) -> u32 {
-        unsafe {
-            let text_len = (self.data.as_ptr() as *const u32).add(1).read_unaligned();
-            text_len + 8
-        }
+        let text_len = unsafe { (self.data.as_ptr() as *const u32).add(1).read_unaligned() };
+        text_len + 8
     }
 
     pub(crate) fn id(&self) -> u32 {
@@ -327,8 +360,57 @@ impl<K: SyntaxKind> SyntaxTree<K> {
     }
 }
 
-impl<K: SyntaxKind> std::fmt::Debug for SyntaxTree<K> {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+struct Events<'a, K> {
+    idx: u32,
+    tree: &'a SyntaxTree<K>,
+}
+
+impl<K: SyntaxKind> Iterator for Events<'_, K> {
+    type Item = Event<K>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        if self.idx >= self.tree.data.len() as u32 {
+            return None;
+        }
+
+        if unsafe { self.tree.is_start_node(self.idx) } {
+            let (kind, _, start, end) = unsafe { self.tree.get_start_node(self.idx) };
+            let range = TextRange::new(start.into(), end.into());
+            self.idx += START_NODE_SIZE;
+            return Some(Event::StartNode { kind, range });
+        }
+
+        if unsafe { self.tree.is_add_token(self.idx) } {
+            let (kind, start, end) = unsafe { self.tree.get_add_token(self.idx) };
+            let range = TextRange::new(start.into(), end.into());
+            self.idx += ADD_TOKEN_SIZE;
+            return Some(Event::AddToken { kind, range });
+        }
+
+        if unsafe { self.tree.is_finish_node(self.idx) } {
+            self.idx += FINISH_NODE_SIZE;
+            return Some(Event::FinishNode);
+        }
+
+        unreachable!()
+    }
+}
+
+/// The events in a syntax tree, as emitted by [`SyntaxTree::events`].
+///
+/// All data here is exactly as it is stored in the tree, with nothing extra computed.
+#[derive(Clone, Copy, PartialEq, Eq, Hash)]
+pub enum Event<K> {
+    #[allow(missing_docs)]
+    StartNode { kind: K, range: TextRange },
+    #[allow(missing_docs)]
+    AddToken { kind: K, range: TextRange },
+    #[allow(missing_docs)]
+    FinishNode,
+}
+
+impl<K: SyntaxKind> fmt::Debug for SyntaxTree<K> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         if !f.alternate() {
             return f.debug_struct("SyntaxTree").field("data", &self.data).finish();
         }
@@ -374,6 +456,16 @@ impl<K: SyntaxKind> std::fmt::Debug for SyntaxTree<K> {
     }
 }
 
+impl<K: SyntaxKind> fmt::Debug for Event<K> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::StartNode { kind, range } => write!(f, "START_NODE {kind:?} {range:?}"),
+            Self::AddToken { kind, range } => write!(f, "ADD_TOKEN {kind:?} {range:?}"),
+            Self::FinishNode => write!(f, "FINISH_NODE"),
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -436,6 +528,26 @@ mod tests {
         assert_eq!(tree.data[4..], data);
     }
 
+    fn big_tree() -> SyntaxTree<SyntaxKind> {
+        let mut builder = SyntaxBuilder::new("# foo\nfncbar->{};");
+
+        builder.start_node(SyntaxKind::Root);
+        builder.add_token(SyntaxKind::Comment, TextRange::new(0.into(), 6.into()));
+        builder.start_node(SyntaxKind::Function);
+        builder.add_token(SyntaxKind::FncKw, TextRange::new(6.into(), 9.into()));
+        builder.add_token(SyntaxKind::Ident, TextRange::new(9.into(), 12.into()));
+        builder.add_token(SyntaxKind::Arrow, TextRange::new(12.into(), 14.into()));
+        builder.start_node(SyntaxKind::Block);
+        builder.add_token(SyntaxKind::LBrace, TextRange::new(14.into(), 15.into()));
+        builder.add_token(SyntaxKind::RBrace, TextRange::new(15.into(), 16.into()));
+        builder.finish_node();
+        builder.add_token(SyntaxKind::Semicolon, TextRange::new(16.into(), 17.into()));
+        builder.finish_node();
+        builder.finish_node();
+
+        builder.finish()
+    }
+
     #[test]
     fn just_root() {
         check(
@@ -494,22 +606,6 @@ mod tests {
 
     #[test]
     fn debug_complex() {
-        let mut builder = SyntaxBuilder::new("# foo\nfncbar->{};");
-        builder.start_node(SyntaxKind::Root);
-        builder.add_token(SyntaxKind::Comment, TextRange::new(0.into(), 6.into()));
-        builder.start_node(SyntaxKind::Function);
-        builder.add_token(SyntaxKind::FncKw, TextRange::new(6.into(), 9.into()));
-        builder.add_token(SyntaxKind::Ident, TextRange::new(9.into(), 12.into()));
-        builder.add_token(SyntaxKind::Arrow, TextRange::new(12.into(), 14.into()));
-        builder.start_node(SyntaxKind::Block);
-        builder.add_token(SyntaxKind::LBrace, TextRange::new(14.into(), 15.into()));
-        builder.add_token(SyntaxKind::RBrace, TextRange::new(15.into(), 16.into()));
-        builder.finish_node();
-        builder.add_token(SyntaxKind::Semicolon, TextRange::new(16.into(), 17.into()));
-        builder.finish_node();
-        builder.finish_node();
-
-        let tree = builder.finish();
         expect![[r##"
             Root@0..17
               Comment@0..6 "# foo\n"
@@ -522,7 +618,29 @@ mod tests {
                   RBrace@15..16 "}"
                 Semicolon@16..17 ";"
         "##]]
-        .assert_eq(&format!("{tree:#?}"));
+        .assert_eq(&format!("{:#?}", big_tree()));
+    }
+
+    #[test]
+    fn events() {
+        expect![[r#"
+            [
+                START_NODE Root 0..17,
+                ADD_TOKEN Comment 0..6,
+                START_NODE Function 6..17,
+                ADD_TOKEN FncKw 6..9,
+                ADD_TOKEN Ident 9..12,
+                ADD_TOKEN Arrow 12..14,
+                START_NODE Block 14..16,
+                ADD_TOKEN LBrace 14..15,
+                ADD_TOKEN RBrace 15..16,
+                FINISH_NODE,
+                ADD_TOKEN Semicolon 16..17,
+                FINISH_NODE,
+                FINISH_NODE,
+            ]
+        "#]]
+        .assert_debug_eq(&big_tree().events().collect::<Vec<_>>());
     }
 
     #[test]
